"use strict";
// Copyright (c) Microsoft Corporation. All rights reserved. Licensed under the MIT license.
// See LICENSE in the project root for license information.
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
const path_1 = __importDefault(require("path"));
const node_core_library_1 = require("@rushstack/node-core-library");
const regexpUtils_1 = require("./regexpUtils");
const PLUGIN_NAME = 'EmbeddedDependenciesWebpackPlugin';
const PLUGIN_ERROR_PREFIX = '[embedded-dependencies-webpack-plugin]';
const DEFAULT_GENERATED_LICENSE_FILE_NAME = 'THIRD-PARTY-NOTICES.html';
const DEFAULT_EMBEDDED_DEPENDENCIES_FILE_NAME = 'embedded-dependencies.json';
const DEFAULT_PACKAGE_FILTER_FUNCTION = () => true;
/**
 * @beta
 * Webpack plugin that generates a file with the list of embedded dependencies
 * and their licenses.
 */
class EmbeddedDependenciesWebpackPlugin {
    constructor(options) {
        this._outputFileName = (options === null || options === void 0 ? void 0 : options.outputFileName) || DEFAULT_EMBEDDED_DEPENDENCIES_FILE_NAME;
        this._generateLicenseFile = (options === null || options === void 0 ? void 0 : options.generateLicenseFile) || false;
        this._generateLicenseFileFunction =
            (options === null || options === void 0 ? void 0 : options.generateLicenseFileFunction) || this._defaultLicenseFileGenerator;
        this._generatedLicenseFilename = (options === null || options === void 0 ? void 0 : options.generatedLicenseFilename) || DEFAULT_GENERATED_LICENSE_FILE_NAME;
        this._packageFilterFunction = (options === null || options === void 0 ? void 0 : options.packageFilterPredicate) || DEFAULT_PACKAGE_FILTER_FUNCTION;
    }
    /**
     * @beta
     * Webpack plugin apply method. This method is called by the webpack compiler to apply the plugin, however it not usually
     * needed to be invoked manually by the developer in a webpack configuration. However, if you are calling this plugin (applying it from another plugin)
     * you can call `plugin.apply(compiler)` to apply the plugin and invoke it.
     * @param compiler - The webpack compiler instance.
     */
    apply(compiler) {
        const { sources, Compilation } = compiler.webpack;
        // Tap into compilation so we can tap into compilation.hooks.processAssets
        compiler.hooks.thisCompilation.tap(PLUGIN_NAME, (compilation, { normalModuleFactory }) => {
            const thirdPartyPackages = new Map();
            normalModuleFactory.hooks.module.tap(PLUGIN_NAME, (module, moduleCreateData, resolveData) => {
                const { resourceResolveData } = moduleCreateData;
                const pkg = resourceResolveData === null || resourceResolveData === void 0 ? void 0 : resourceResolveData.descriptionFileData;
                const filePath = resourceResolveData === null || resourceResolveData === void 0 ? void 0 : resourceResolveData.descriptionFileRoot;
                if (pkg &&
                    filePath &&
                    this._packageFilterFunction(pkg, filePath) &&
                    (filePath === null || filePath === void 0 ? void 0 : filePath.includes('node_modules'))) {
                    const key = makePackageMapKeyForPackage(pkg);
                    thirdPartyPackages.set(key, { packageFolderPath: filePath, packageJsonData: pkg });
                }
                return module;
            });
            compilation.hooks.processAssets.tapPromise({ name: PLUGIN_NAME, stage: Compilation.PROCESS_ASSETS_STAGE_REPORT }, async (assets) => {
                const packages = [];
                try {
                    await node_core_library_1.Async.forEachAsync(thirdPartyPackages, async ([, { packageFolderPath: dir, packageJsonData: data }]) => {
                        const { name, version } = data;
                        let licenseSource;
                        const license = parseLicense(data);
                        const licensePath = await this._getLicenseFilePathAsync(dir, compiler);
                        if (licensePath) {
                            licenseSource = await node_core_library_1.FileSystem.readFileAsync(licensePath);
                            const copyright = this._parseCopyright(licenseSource) || parsePackageAuthor(data);
                            packages.push({
                                name,
                                version,
                                license,
                                licenseSource,
                                copyright
                            });
                        }
                        else {
                            // If there is no license file path, we still should populate the other required fields
                            const copyright = parsePackageAuthor(data);
                            packages.push({
                                name,
                                version,
                                license,
                                copyright
                            });
                        }
                    });
                }
                catch (error) {
                    this._emitWebpackError(compilation, 'Failed to process embedded dependencies', error);
                }
                finally {
                    node_core_library_1.Sort.sortBy(packages, (pkg) => pkg.name);
                }
                const dataToStringify = {
                    embeddedDependencies: packages
                };
                compilation.emitAsset(this._outputFileName, new sources.RawSource(JSON.stringify(dataToStringify)));
                if (this._generateLicenseFile) {
                    // We should try catch here because generator function can be output from user config
                    try {
                        compilation.emitAsset(this._generatedLicenseFilename, new sources.RawSource(this._generateLicenseFileFunction(packages)));
                    }
                    catch (error) {
                        this._emitWebpackError(compilation, 'Failed to generate license file', error);
                    }
                }
                return;
            });
        });
    }
    /**
     * Default error handler for try/catch blocks in the plugin
     * try/catches emit errors of type `unknown` and we need to handle them based on what
     * type the error is. This function provides a convenient way to handle errors and then
     * propagate them to webpack as WebpackError objects on `compilation.errors` array.
     *
     * @remarks
     * _If we need to push errors to `compilation.warnings` array, we should just create a companion function
     * that does the same thing but pushes to `compilation.warnings` array instead._
     *
     * @example
     * ```typescript
     * try {
     *   // do some operation
     *   FileSystem.readFile('some-file');
     * } catch (error: unknown) {
     *   this._emitWebpackError(compilation, 'Failed to do some operation', error);
     * }
     * ```
     */
    _emitWebpackError(compilation, errorMessage, error) {
        let emittedError;
        const { WebpackError } = compilation.compiler.webpack;
        // If the error is a string, we can just emit it as is with message prefix and error message
        if (typeof error === 'string') {
            emittedError = new WebpackError(`${PLUGIN_ERROR_PREFIX}: ${errorMessage}: ${error}`);
            // If error is an instance of Error, we can emit it with message prefix, error message and stack trace
        }
        else if (error instanceof Error) {
            emittedError = new WebpackError(`${PLUGIN_ERROR_PREFIX}: ${errorMessage}: ${error.message}\n${error.stack || ''}`);
            // If error is not a string or an instance of Error, we can emit it with message prefix and error message and JSON.stringify it
        }
        else {
            emittedError = new WebpackError(`${PLUGIN_ERROR_PREFIX}: ${errorMessage}: ${JSON.stringify(error || '')}`);
        }
        compilation.errors.push(emittedError);
    }
    /**
     * Searches a third party package directory for a license file.
     */
    async _getLicenseFilePathAsync(modulePath, compiler) {
        // TODO: Real fs.readdir can take an arguement ({ withFileTypes: true }) which will filter out directories for better performance
        //       and return a list of Dirent objects. Currently the webpack types are hand generated for fs.readdir so
        //       we can't use this feature yet, or we would have to cast the types of inputFileSystem.readdir.
        //       https://github.com/webpack/webpack/issues/16780 tracks this issue.
        const files = await node_core_library_1.LegacyAdapters.convertCallbackToPromise(compiler.inputFileSystem.readdir, modulePath);
        return files === null || files === void 0 ? void 0 : files.map((file) => file.toString()).filter((file) => regexpUtils_1.LICENSE_FILES_REGEXP.test(file)).map((file) => path_1.default.join(modulePath, file))[0]; // Grabbing the first license file if multiple are found
    }
    /**
     * Given a module path, try to parse the module's copyright attribution.
     */
    _parseCopyright(licenseSource) {
        const match = licenseSource.match(regexpUtils_1.COPYRIGHT_REGEX);
        if (match) {
            return match[0];
        }
        return undefined;
    }
    _defaultLicenseFileGenerator(packages) {
        const licenseContent = (pkg) => pkg.licenseSource || pkg.copyright || 'License or Copyright not found';
        const licenseTemplateForPackage = (pkg) => {
            return `<hr />${pkg.name} - ${pkg.version}<br /><br />${licenseContent(pkg)}`;
        };
        return packages.map(licenseTemplateForPackage).join('\n');
    }
}
exports.default = EmbeddedDependenciesWebpackPlugin;
function makePackageMapKeyForPackage(pkg) {
    return `${pkg.name}@${pkg.version}`;
}
/**
 * Returns the license type
 */
function parseLicense(packageData) {
    var _a;
    if (packageData.license) {
        return packageData.license;
    }
    else if (typeof packageData.licenses === 'string') {
        return packageData.licenses;
    }
    else if ((_a = packageData.licenses) === null || _a === void 0 ? void 0 : _a.length) {
        return packageData.licenses.length === 1
            ? packageData.licenses[0].type
            : `(${packageData.licenses
                .map((license) => license.type)
                .join(' OR ')})`;
    }
    return undefined;
}
function parsePackageAuthor(p) {
    var _a;
    return typeof p.author === 'string' ? p.author : (_a = p.author) === null || _a === void 0 ? void 0 : _a.name;
}
//# sourceMappingURL=EmbeddedDependenciesWebpackPlugin.js.map